//----------------------------------------------------------------------------------------------
// BoardBoard Library : Silicon Labs 2.3.0
// Board Select       : Seeed Studio XIAO MG24 (Sense)
// Plotocol stack     : BLE(Arduino) or BLE(Silabs) or None
//----------------------------------------------------------------------------------------------
// [NOTE] To measure battery voltage correctly, set Vref=1.21
//        \Arduino15\packages\SiliconLabs\hardware\silabs\2.3.0\cores\silabs\adc.cpp
//        Add LINE 93   all_configs.configs[0].analogGain = iadcCfgAnalogGain0P5x;
// [NOTE] “uint32_t timestamp = millis();” will hang ----> uint32_t timestamp;

// 1. wait for a packet from Slave
// 2. receives Slave data 11 bytes(SlaveUID 4, randomNumber 4, Vbatt 2, (previous)veryfiResult 1)
// 3. transtits response data 11 bytes(MasterUID 4, verifyNumber 4, rssi 1, snr 1, interval 1)

// 2025/04/04

#include <RadioLib.h>           // 7.1.2  RadioLib by Jan Gromes
#include <U8g2lib.h>            // U8g2 2.35.30
#include <SPI.h>
#include <SD.h>

#define CHECK_INTERVAL 15000    // communication check interval [mS]
#define RF_SW          D5       // PC5 RF Switch
#define SD_CS          D0       // PC0 SD chip select pin
#define VBAT          (24)      // PD4 battery voltage measurement ADC pin
#define VBAT_CTL      (25)      // PD3 read VBAT enable HIGH:enable, LOW:disable

// getDeviceUniqueId() to find the UID
#define SLAVE_UID   0xFE55C37C  // Slave device UID example
#define MASTER_UID  0xFE55C38A  // Master device UID example

// SSD1306 software I2C library, SCL=D6, SDA=D7, 400kHz
U8G2_SSD1306_128X64_NONAME_F_SW_I2C display(U8G2_R0, /*clock=*/ D6, /*data=*/ D7, /*reset=*/ U8X8_PIN_NONE);
// SD log file
File logFile;
// Radio lib Modele default setting : SPI:2MHz, MSBFIRST, MODE0
SX1262 radio = new Module(/*NSS*/D4, /*DIO1*/D1, /*NRST*/D2, /*BUSY*/D3);

#define dataNum   11                // receive data 11 bytes  (SlaveUID 4, randomNumber 4, Vbatt 2, (previous)veryfiResult 1)
                                    // transmit data 11 bytes (MasterUID 4, verifyNumber 4, rssi 1, snr 1, interval 1)
union unionBuff {                   // buffer for data type conversion
  uint32_t  Buff_u32[dataNum/4];    // SlaveID, MasterID, randomNumber, verifyNumber
  uint16_t  Buff_u16[dataNum/2];    // Vbatt
  int8_t    Buff_i8[dataNum];       // rssi, snr
  uint8_t   Buff_u8[dataNum];       // verifyResult, interval
};
union unionBuff ub;

int8_t rssi;                        // signal RSSI [dBm]
int8_t snr;                         // signal SN ratio [dB]
uint32_t masteruid = MASTER_UID;    // Master device UID
uint32_t slaveuid;                  // Slave device UID
uint16_t Vbatt;                     // battery voltage data [V]
uint32_t randomNumber;              // random number to verify
uint32_t verifyNumber;              // verify number
uint8_t verifyResult;               // received verify result
char printBuff[4];                  // for sprintf()
String  txdata = "";                // transmission data packet string
String  rxdata = "";                // received data packet string
bool receivedFlag = false;          // flag that a packet was received
bool transmittedFlag = false;       // flag that a packet was transmitted
bool operationDone = false;         // receive or transmit operation done
String logFileName = "LoRaLog.txt"; // Log file name
uint32_t timestamp;                 // previouse received time [mS]
uint32_t receivedTime;              // packet received time [mS]
uint8_t receivedInterval;           // packet received interval [sec]
int state;                          // state of radio module
uint32_t timeoutCheck;              // for timeout check

// *******************************************************************************************************
void setup() {
  Serial.begin(115200);
  
  // Master UID
  Serial.print("Master UID "); Serial.println(MASTER_UID, HEX);

  // Pin initialization
  pinMode(RF_SW, OUTPUT);               // RF Switch
  pinMode(LED_BUILTIN, OUTPUT);         // built-in LED
  pinMode(VBAT_CTL, OUTPUT);            // VBAT read enable
  digitalWrite(RF_SW, HIGH);            // internal DIO2 controls Tx/Rx
  digitalWrite(LED_BUILTIN, HIGH);      // LED off
  digitalWrite(VBAT_CTL, HIGH);         // read enable

  // ADC initialization
  // To measure battery voltage correctly, set Vref=1.21
  // \Arduino15\packages\SiliconLabs\hardware\silabs\2.3.0\cores\silabs\adc.cpp
  // Add LINE 93   all_configs.configs[0].analogGain = iadcCfgAnalogGain0P5x;
  ADC.set_reference(AR_INTERNAL1V2);    // internal 1.21V   See adc.h enum analog_references{};
  ADC.set_read_resolution(12);          // 12 bit

  // SSD1306 Display initialization
  display.begin();
  display.clearDisplay();
  display.setFont(u8g2_font_8x13B_tr);
  display.setCursor(0, 15);
  display.println("Wio-SX1262");
  display.setCursor(0, 31);
  display.println("P2P Master");
  display.sendBuffer();
  delay(1000);
  display.clearDisplay();

  // int16_t begin(float freq = 434.0, float bw = 125.0, uint8_t sf = 9, uint8_t cr = 7, \
  // uint8_t syncWord = RADIOLIB_SX126X_SYNC_WORD_PRIVATE, int8_t power = 10, \
  // uint16_t preambleLength = 8, float tcxoVoltage = 1.6, bool useRegulatorLDO = false);
  // [NOTE} To be set up in compliance with area rules
  int state = radio.begin(922.6, 125.0, 12, 5, 0x12, 13, 8, 1.7, 0);

  if (state == RADIOLIB_ERR_NONE) {
    Serial.println("radio.begin() success!");
  } else {
    Serial.print("failed, code ");
    Serial.println(state);
    errorBlink(4);                           // error [[ 4 ]]
  }

  // callback function when receive or transmit operation done
  radio.setDio1Action(setFlag);

  // SD initialization and file open
  if(!SD.begin(2000000UL, SD_CS)) {   
    Serial.println("SD begin failure!");
    display.clearDisplay();
    display.setCursor(0, 15);
    display.println("SD begin failure");
    display.sendBuffer();
    errorBlink(2);                           // error [[ 2 ]]
  } 

  logFile = SD.open(logFileName, FILE_WRITE);
  if(!logFile) { 
    Serial.println(logFileName + " open failure");
    display.clearDisplay();
    display.setCursor(0, 15);
    display.println("file open failure");
    display.sendBuffer();
    errorBlink(3);                           // error [[ 3 ]]   
  }   
  Serial.println("SD initialization finished");

  // setup end
  display.setCursor(0, 15);
  display.print("-- Receiving --");
  display.sendBuffer();
}

// ***********************************************************************************************************
void loop() 
{
  Serial.println("******** Waiting for incoming Slave transmitting packet ********");

  // start listening for LoRa packets
  digitalWrite(LED_BUILTIN, LOW);
      operationDone = false;    
      state = radio.startReceive();
      // check if the flag is set
      // If not received, proceed to next communication cycle to atempt synchronization recovery
      timeoutCheck = millis();
      while(!operationDone & ((millis() - timeoutCheck) < CHECK_INTERVAL)) {
        delay(1);
      }   
  digitalWrite(LED_BUILTIN, HIGH);

  // received status check  
  if (state != RADIOLIB_ERR_NONE) {
    Serial.print("failed code ");
    Serial.println(state);
    errorBlink_1(2);                        // error [[ 2 ]]
  }

  // read a packet 
  state = radio.readData(ub.Buff_u8, dataNum);

  slaveuid = ub.Buff_u32[0];    // Slave device UID
  if (state == RADIOLIB_ERR_NONE & slaveuid == SLAVE_UID) {   // Check Slave ID   
    receivedTime = millis();
    receivedInterval = (receivedTime - timestamp) / 1000;       // [sec]
    timestamp = receivedTime;

    rxdata = "";
    for(int i = 0; i < dataNum; i++) {
      sprintf(printBuff, "%02x", ub.Buff_u8[i]);
      rxdata += printBuff;
    }
    rxdata.toUpperCase();
    Serial.println(rxdata);

    // restoration of received data
    randomNumber = ub.Buff_u32[1];    // random number to verify
    Vbatt = ub.Buff_u16[4];           // battery voltage [mV]
    verifyResult = ub.Buff_u8[10];    // verify result 

    Serial.printf("Slave UID:\t%08x\n", slaveuid);    
    Serial.print("Interval:\t"); Serial.print(receivedInterval); Serial.println(" sec");
    Serial.print("Vbatt:\t\t"); Serial.print(Vbatt/1000.0); Serial.println(" V");
    Serial.print("randomNumber:\t"); Serial.println(randomNumber, HEX);
    Serial.print("Verify:\t\t"); Serial.println(verifyResult ? "good" : "bad");

    // RSSI and SNR of the last received packet
    rssi = radio.getRSSI();
    Serial.print("RSSI:\t\t"); Serial.print(rssi); Serial.println(" dBm");
    snr = radio.getSNR();
    Serial.print("SNR:\t\t"); Serial.print(snr); Serial.println(" dB");

    display.clearDisplay();
    display.setCursor(0, 15); display.print("Received"); 
    display.setCursor(0, 31); display.print(randomNumber, HEX);
    display.setCursor(80, 31); display.print(verifyResult ? "good" : "bad");    
    display.setCursor(0, 47); display.print(rssi); display.print(" dBm");
    display.setCursor(80, 47); display.print(snr); display.print(" dB");
    display.setCursor(0, 63); display.print(receivedTime / 1000); display.print(" sec");
    display.setCursor(80, 63); display.print(Vbatt/1000.0); display.print(" V");   
    display.sendBuffer();
    
    // write to SD
    logFile.print(receivedTime); logFile.print(",");
    logFile.print(receivedInterval); logFile.print(",");
    logFile.print(rxdata); logFile.print(",");
    logFile.print(slaveuid); logFile.print(",");
    logFile.print(verifyResult); logFile.print(",");    
    logFile.print(Vbatt); logFile.print(",");
    logFile.print(rssi); logFile.print(",");
    logFile.println(snr);
    logFile.flush();
  } else if (state == RADIOLIB_ERR_CRC_MISMATCH) {
    // packet was received, but is malformed
    Serial.println("******** CRC error ********");
  } else {
    // some other error occurred when receiving
    Serial.print("receive failed, code ");
    Serial.println(state);
    errorBlink_1(3);                        // error [[ 3 ]]                  
  }
  digitalWrite(LED_BUILTIN, HIGH);
  delay(1000);     


//----------------------------------------------------------------------------------------------
  Serial.println("******** Transmitting response packet ********");

  // battery voltage measurement
  int n = 256;
  uint32_t tempVadc = 0;
  for(int i = 0; i < n; i++) {
    tempVadc = tempVadc + analogRead(VBAT);
  }
  uint16_t Vadc = tempVadc / n;
  Vbatt = (2 * 2 * 1210 * Vadc) / 4096;   // gain=0.5, devided 1/2, Vref=1210mV, 12bits

  // transmit data setting
  ub.Buff_u32[0] = masteruid;         // Master device UID
  verifyNumber = randomNumber;
  ub.Buff_u32[1] = verifyNumber;      // to verify
  ub.Buff_i8[8] = rssi;               // Master RSSI  [dBm]
  ub.Buff_i8[9] = snr;                // Master SN ratio  [dB]
  ub.Buff_u8[10] = receivedInterval;  // packet reserved interval [sec]

  digitalWrite(LED_BUILTIN, LOW);
      operationDone = false;
      state = radio.startTransmit(ub.Buff_u8, dataNum);

      // wait for transmittion completion
      timeoutCheck = millis();
      while(!operationDone & ((millis() - timeoutCheck) < CHECK_INTERVAL)) {
        delay(1);
      }
  digitalWrite(LED_BUILTIN, HIGH);

  // print and display transmitted data
  txdata = "";
  for(int i = 0; i < dataNum; i++) {
    sprintf(printBuff, "%02x", ub.Buff_u8[i]);
    txdata += printBuff;
  }
  txdata.toUpperCase();
  Serial.println(txdata);

  Serial.printf("Master UID:\t%08x\n", masteruid);
  Serial.print("RSSI\t\t"); Serial.print(rssi); Serial.println(" dBm");
  Serial.print("SNR\t\t"); Serial.print(snr); Serial.println(" dB");
  Serial.print("verifyNumber:\t"); Serial.println(verifyNumber, HEX);

  display.clearDisplay();
  display.setCursor(40, 15); display.print("Transmitted");  
  display.setCursor(0, 31); display.print(verifyNumber, HEX);
  display.setCursor(80, 31); display.print(verifyResult ? "good" : "bad");
  display.setCursor(0, 47); display.print(rssi); display.print(" dBm");
  display.setCursor(80, 47); display.print(snr); display.print(" dB");
  display.setCursor(0, 63); display.print(receivedInterval); display.print(" sec");
  display.setCursor(80, 63); display.print(Vbatt/1000.0); display.print(" V");
  display.sendBuffer();   

  // error status check
  if (state == RADIOLIB_ERR_NONE) {
    // packet was successfully sent
    Serial.println("transmission finished!");
    Serial.println();
  } 
  else {
    // some other error occurred
    Serial.print("failed, code ");
    Serial.println(state);
  }
  digitalWrite(LED_BUILTIN, HIGH);
  delay(1000);   
}

//*********************************************************************************
// this function is called when a complete packet is received or transmitted
void setFlag(void) {
  operationDone = true;
}


//*********************************************************************************
// Blinks specified number of times on error
void errorBlink(uint8_t err)
{
  while(true) {
    for(int i = 0; i < err; i++) {
      digitalWrite(LED_BUILTIN, LOW);
      delay(200);
      digitalWrite(LED_BUILTIN, HIGH);
      delay(200);
    }
  delay(500);
  }
}

void errorBlink_1(uint8_t err)
{
  for(int i = 0; i < err; i++) {
    digitalWrite(LED_BUILTIN, LOW);
    delay(200);
    digitalWrite(LED_BUILTIN, HIGH);
    delay(200);
  }
  delay(500);
}