//----------------------------------------------------------------------------------------------
// BoardBoard Library : Silicon Labs 2.3.0
// Board Select       : Seeed Studio XIAO MG24 (Sense)
// Plotocol stack     : BLE(Arduino) or BLE(Silabs) or None
//----------------------------------------------------------------------------------------------
// [NOTE] To measure battery voltage correctly, set Vref=1.21
//        \Arduino15\packages\SiliconLabs\hardware\silabs\2.3.0\cores\silabs\adc.cpp
//        Add LINE 93   all_configs.configs[0].analogGain = iadcCfgAnalogGain0P5x;
// [NOTE] “uint32_t timestamp = millis();” will hang ----> uint32_t timestamp;

// 1. transmit data 11 bytes(SlaveUID 4, randomNumber 4, Vbatt 2, (previous)veryfiResult 1)
// 2. wait for response from Master
// 3. receive data 11 bytes(MasterUID 4, verifyNumber 4, rssi 1, snr 1, interval 1)
// 4. verify sent randomNumber with received verifyNumber

// 2025/04/04

#include <RadioLib.h>           // 7.1.2  RadioLib by Jan Gromes
#include <U8g2lib.h>            // U8g2 2.35.30

#define CHECK_INTERVAL  15000   // communication check interval [mS]
#define Tx_INTERVAL     10000   // depend on the erea rules, total of 360 sec or less per hour
#define VBAT           (24)     // PD4 battery voltage measurement ADC pin
#define VBAT_CTL       (25)     // PD3 read VBAT enable HIGH:enable, LOW:disable
#define RF_SW           D5      // PC5 RF Switch

// getDeviceUniqueId() to find the UID
#define SLAVE_UID   0xFE55C37C  // Slave device UID example
#define MASTER_UID  0xFE55C38A  // Master device UID example

// SSD1306 software I2C library, SCL=D6, SDA=D7, 400kHz
U8G2_SSD1306_128X64_NONAME_F_SW_I2C display(U8G2_R0, /*clock=*/ D6, /*data=*/ D7, /*reset=*/ U8X8_PIN_NONE);
// Radio lib Modele default setting : SPI:2MHz, MSBFIRST, MODE0
SX1262 radio = new Module(/*NSS*/D4, /*DIO1*/D1, /*NRST*/D2, /*BUSY*/D3);

#define dataNum   11                // transmit data 11 bytes (SlaveUID 4, randomNumber 4, Vbatt 2, (previous)veryfiResult 1)
                                    // receive data 11 bytes  (MasterUID 4, verifyNumber 4, rssi 1, snr 1, interval 1)
union unionBuff {                   // buffer for data type conversion
  uint32_t  Buff_u32[dataNum/4];    // SlaveID, MasterID, randomNumber, verifyNumber
  uint16_t  Buff_u16[dataNum/2];    // Vbatt
  int8_t    Buff_i8[dataNum];       // rssi, snr
  uint8_t   Buff_u8[dataNum];       // verifyResult, interval
};
union unionBuff ub;

int8_t rssi;                        // signal RSSI [dBm]
int8_t snr;                         // signal SN ratio [dB]
uint32_t slaveuid = SLAVE_UID;      // Slave device UID
uint32_t masteruid;                 // Master device UID
uint16_t Vbatt;                     // battery voltage data [mV]
char printBuff[4];                  // for sprintf()
String  txdata = "";                // transmission data packet string
String  rxdata = "";                // received data packet string
bool receivedFlag = false;          // flag that a packet was received
bool transmittedFlag = false;       // flag that a packet was transmitted
bool operationDone = false;         // receive or transmit operation done
uint8_t receivedInterval;           // Master reveived interval [sec]
uint32_t randomNumber;              // random number to verify that data was sent correctly
uint32_t verifyNumber;              // verify number from Master
uint32_t verifyResult = 0;          // verify result 1:GOOD or 0:ERROR
int state;                          // state of radio module
uint32_t timeoutCheck;              // for timeout check
uint32_t txtime;                    // transmission time [mS]

// *******************************************************************************************************
void setup() {
  Serial.begin(115200);

  // Slave UID
  Serial.print("Slave UID "); Serial.println(SLAVE_UID, HEX);

  // Pin initialization
  pinMode(RF_SW, OUTPUT);               // RF Switch
  pinMode(LED_BUILTIN, OUTPUT);         // built-in LED
  pinMode(VBAT_CTL, OUTPUT);            // VBAT read enable
  digitalWrite(RF_SW, HIGH);            // internal DIO2 controls Tx/Rx
  digitalWrite(LED_BUILTIN, HIGH);      // LED off  
  digitalWrite(VBAT_CTL, HIGH);         // enable

  // ADC initialization
  // To measure battery voltage correctly, set Vref=1.21
  // \Arduino15\packages\SiliconLabs\hardware\silabs\2.3.0\cores\silabs\adc.cpp
  // Add LINE 93   all_configs.configs[0].analogGain = iadcCfgAnalogGain0P5x;
  ADC.set_reference(AR_INTERNAL1V2);    // internal 1.21V   See adc.h enum analog_references{};
  ADC.set_read_resolution(12);          // 12 bit

  // Display initialization
  display.begin();
  display.clearDisplay();
  display.setFont(u8g2_font_8x13B_tr);
  display.setCursor(0, 15);
  display.println("Wio-SX1262");
  display.setCursor(0, 31);
  display.println("P2P Slave");
  display.sendBuffer();
  delay(1000);
  display.clearDisplay();  


  // Radio initialization, depend on the erea rules
  // int16_t begin(float freq = 434.0, float bw = 125.0, uint8_t sf = 9, uint8_t cr = 7, \
  // uint8_t syncWord = RADIOLIB_SX126X_SYNC_WORD_PRIVATE, int8_t power = 10, \
  // uint16_t preambleLength = 8, float tcxoVoltage = 1.6, bool useRegulatorLDO = false);
  // [NOTE] to be set up in compliance with area rules
  state = radio.begin(922.6, 125.0, 12, 5, 0x12, 13, 8, 1.7, 0);

  if (state == RADIOLIB_ERR_NONE) {
    Serial.println("radio.begin() success!");
  }
  else {
    Serial.print("failed, code ");
    Serial.println(state);
    errorBlink(4);                // error [[ 4 ]]
  }

  // callback function when receive or transmit operation done
  radio.setDio1Action(setFlag);

  // setup end
  display.setCursor(0, 63);
  display.print("Setup END");
  display.sendBuffer();
  delay(1000);
  display.clearDisplay();
}

// ***********************************************************************************************************
void loop() 
{
  do {  
    Serial.println("******** Slave transmitting packet ********");

    // battery voltage measurement
    int n = 256;
    uint32_t tempVadc = 0;
    for(int i = 0; i < n; i++) {
      tempVadc = tempVadc + analogRead(VBAT);
    }
    uint16_t Vadc = tempVadc / n;
    Vbatt = (2 * 2 * 1210 * Vadc) / 4096;   // gain=0.5, devided 1/2, Vref=1210mV, 12bits

    // transmit data setting
    ub.Buff_u32[0] = slaveuid;                      // Slave device UID
    randomNumber = random(0x7FFFFFFF);              // for verification
    ub.Buff_u32[1] = randomNumber;
    ub.Buff_u16[4] = Vbatt;                         // battery voltage [mV]
    ub.Buff_u8[10] = verifyResult;                  // previous verify result

    // transmit a packet
    txtime = millis(); 
    uint16_t sec = txtime / 1000;                   // transmission time [sec] 

    digitalWrite(LED_BUILTIN, LOW);
        operationDone = false;
        state = radio.startTransmit(ub.Buff_u8, dataNum);

        // wait for transmittion completion
        timeoutCheck = millis();
        while(!operationDone & ((millis() - timeoutCheck) < CHECK_INTERVAL)) {
          delay(1);
        }
    digitalWrite(LED_BUILTIN, HIGH);
   
    // print and display transmitted data
    txdata = "";
    for(int i = 0; i < dataNum; i++) {
      sprintf(printBuff, "%02x", ub.Buff_u8[i]);
      txdata += printBuff;
    }
    txdata.toUpperCase();
    Serial.println(txdata);

    Serial.print("Slave UID:\t"); Serial.println(slaveuid, HEX);  
    Serial.print("Time:\t\t"); Serial.print(sec); Serial.println(" sec");
    Serial.print("Vbatt:\t\t"); Serial.print(Vbatt/1000.0); Serial.println(" V");
    Serial.print("randomNumber:\t"); Serial.println(randomNumber, HEX);
    Serial.print("VerifyResult:\t"); Serial.println(verifyResult ? "good" : "bad");

    display.clearDisplay();
    display.setCursor(0, 15); display.print("Transmitted");    
    display.setCursor(0, 31); display.print(randomNumber, HEX);
    display.setCursor(0, 63); display.print(sec); display.print(" sec");
    display.setCursor(80, 63); display.print(Vbatt/1000.0); display.print(" V");   
    display.sendBuffer();    
  } while(state != RADIOLIB_ERR_NONE);  // if status error, transmit again

//-------------------------------------------------------------------------------------------------
  Serial.println("******** Waiting for incoming Master response ******** ");

  // start listening for LoRa packets
  digitalWrite(LED_BUILTIN, LOW);
      operationDone = false;    
      state = radio.startReceive();

      // check the flag
      // If not received, proceed to next communication cycle to atempt synchronization recovery
      timeoutCheck = millis();
      while(!operationDone & ((millis() - timeoutCheck) < CHECK_INTERVAL)) {
        delay(1);
      }  
  digitalWrite(LED_BUILTIN, HIGH);

  // received status check  
  if (state != RADIOLIB_ERR_NONE) {
    Serial.print("failed code ");
    Serial.println(state);
    errorBlink_1(2);                        // error [[ 2 ]]
  }

  // read a received packet 
  state = radio.readData(ub.Buff_u8, dataNum);

  // read status and Master UID check
  masteruid = ub.Buff_u32[0];
  if (state == RADIOLIB_ERR_NONE & masteruid == MASTER_UID) {   // chack status and Master UID
    rxdata = "";
    for(int i = 0; i < dataNum; i++) {
      sprintf(printBuff, "%02x", ub.Buff_u8[i]);
      rxdata += printBuff;
    }
    rxdata.toUpperCase();
    Serial.println(rxdata);

    // restoration of received data
    verifyNumber = ub.Buff_u32[1];
    rssi = ub.Buff_i8[8];  // RSSI   min. -127[dBm]
    snr = ub.Buff_i8[9];   // SNR
    receivedInterval = ub.Buff_u8[10];

    // random number verification
    if(randomNumber == verifyNumber) {
      verifyResult = 1;    
    } else {
      verifyResult = 0;
    }
    
    Serial.print("Master UID:\t"); Serial.println(masteruid, HEX);
    Serial.print("Interval:\t"); Serial.print(receivedInterval); Serial.println(" sec");
    Serial.print("RSSI:\t\t"); Serial.print(rssi); Serial.println(" dBm");
    Serial.print("SNR:\t\t"); Serial.print(snr); Serial.println(" dB");
    Serial.print("Vbatt:\t\t"); Serial.print(Vbatt); Serial.println(" mV");
    Serial.print("randomNumber:\t"); Serial.println(randomNumber, HEX);
    Serial.print("verifyNumber:\t"); Serial.println(verifyNumber, HEX);
    Serial.print("Verify\t\t"); Serial.println(verifyResult ? "good" : "bad");
    Serial.println();

    display.clearDisplay();
    display.setCursor(64, 15); display.print("Received");  
    display.setCursor(0, 31); display.print(verifyNumber, HEX);
    display.setCursor(80, 31); display.print(verifyResult ? "good" : "bad");
    display.setCursor(0, 47); display.print(rssi); display.print(" dBm");
    display.setCursor(80, 47); display.print(snr); display.print(" dB");
    display.setCursor(0, 63); display.print(receivedInterval); display.print(" sec");
    display.setCursor(80, 63); display.print(Vbatt/1000.0); display.print(" V");   
    display.sendBuffer();
  }
  else {
    // error occurred when receiving
    Serial.print("receive failed, code ");
    Serial.println(state);
    errorBlink_1(3);                        // error [[ 3 ]]
  }  
  digitalWrite(LED_BUILTIN, HIGH);

  // wait for transmit interval
  while((millis() - txtime) < Tx_INTERVAL);
}

//****************************************************************************************
// this function is called when a complete packet is received or transmitted
void setFlag(void) {
  operationDone = true;
}

//****************************************************************************************
// Blink specified number of times on error
void errorBlink(uint8_t err)
{
  while(true) {
    for(int i = 0; i < err; i++) {
      digitalWrite(LED_BUILTIN, LOW);
      delay(200);
      digitalWrite(LED_BUILTIN, HIGH);
      delay(200);
    }
    delay(500);
  }
}

void errorBlink_1(uint8_t err)
{
  for(int i = 0; i < err; i++) {
    digitalWrite(LED_BUILTIN, LOW);
    delay(200);
    digitalWrite(LED_BUILTIN, HIGH);
    delay(200);
  }
  delay(500);
}